/* Concept and parts of the source borrowed from 
 * "Muli Ben-Yehuda <mulix@mulix.org>"
 */

#include <linux/config.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include <asm/uaccess.h>

#include "klife.h"

#define DEBUG 1

#ifdef DEBUG 
#define DBG(n, ...) printk(n, ##__VA_ARGS__)
#else
#define DBG(n, ...)
#endif

struct klife {
  int grid[GRIDSIZE][GRIDSIZE];
  int tmpgrid[GRIDSIZE][GRIDSIZE];
  int generation;
};


static int klife_open(struct inode *inode, struct file *file) {
  struct klife *k;

  DBG("klife_open called\n");

  k = kmalloc(sizeof(struct klife), GFP_KERNEL);
  if (!k) return -ENOMEM;
  memset(k, 0, sizeof(struct klife));
  
  // je geoeffnetem gibt es ein spielfeld
  file->private_data=k;


  // Initialen Zustand des Spielfeldes setzen
  k->grid[10][10]=1;
  k->grid[11][10]=1;
  k->grid[10][11]=1;
  k->grid[11][11]=1;

  k->generation=0;

  return 0;
}

static int klife_release(struct inode *inode, struct file *file) {
  DBG("klife_release called\n");

  if(file->private_data) 
    kfree(file->private_data);

  return 0;
}

// ermittelt die anzahl der "lebenden" nachbarn eines feldes
static int neighbours(struct klife *k, int x, int y) {
  int i,j, count=0;
  
  for (i=x-1; i<=x+1; i++) {
    if (i < 0 || i >= GRIDSIZE)
      continue;
    for (j=y-1; j<=y+1; j++) {
          if (j < 0 || j >= GRIDSIZE)
	    continue;
	  if (k->grid[i][j]) count++;
    }
  }
  return count;
}

// schreitet zum naechsten spielfeldzustand fort
static void next(struct klife *k) {
  int x,y;

  for (x=0; x<GRIDSIZE; x++)
    for (y=0; y<GRIDSIZE; y++) {
      int count = neighbours(k,x,y);
      if (count > 1 && count < 4) 
	k->tmpgrid[x][y]=1;
      else 
	k->tmpgrid[x][y]=0;

    }
  
  memcpy(k->grid, k->tmpgrid, sizeof(k->grid));

  k->generation++;

  DBG("Advanced to generation %i\n", k->generation);

  return;
}


static ssize_t klife_read(struct file *file, char *buf,
		      size_t count, loff_t *ppos) {
  // buf zeigt auf den vom aufrufenden Prozess bereitgestellten
  // Puffer, count enthaelt dessen Größe und ppos gibt den 
  // Offset innerhalb der Datei an ab dem wir lesen.
  // buf befindet sich im Userspace, wir muessen also zunaechst
  // unseren eigenen Puffer anlegen, und die Daten dann mit 
  // copy_to_user transferieren. 
  
  int x,y;
  struct klife *k = file->private_data;
  char *kbuf=0, *kbuf_orig;
  int textlen = (GRIDSIZE+1)*GRIDSIZE + 1;
  int ret;
  
  DBG("klife_read called\n");
  
  kbuf_orig = kbuf = kmalloc(textlen, GFP_KERNEL);
  if (!kbuf) {
    return -ENOMEM;
  }

  next(k);

  // Wir schreiben der Einfachheit halber nur Daten an den User, 
  // wenn sein Puffer ausreichend gros ist. Der Einfachheit
  // halber.
  if (count < textlen) {
    ret = -ENOMEM;
    goto free;
  }
  
  for (x=0; x<GRIDSIZE; x++) {
    for (y=0; y<GRIDSIZE; y++) {
      *kbuf++ = (k->grid[x][y])?'X':' ';
    }
    *kbuf++='\n';
  }
  *kbuf=0;

  if (copy_to_user(buf, kbuf_orig, textlen)) {
    ret = -EFAULT;
    goto free;
  }

  *ppos+=textlen;  
  ret=textlen;

 free:
  kfree(kbuf_orig);

  return ret;
}

static ssize_t klife_write(struct file *file, const char __user *buf,
		       size_t count, loff_t *ppos){
  DBG("klife_write called\n");
  printk("klife: write not implemented\n");
  return -EINVAL;
}

static int klife_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		       unsigned long arg) {
  struct klife *k = file->private_data;

  DBG("klife_ioctl called\n");

  switch(cmd) {
  case KLIFE_GET_GENERATION:
    copy_to_user((void*)arg, &k->generation, sizeof(&k->generation));
  default:
    printk("klife: unknown ioctl %i\n", cmd);
    return -EINVAL;
  }

  return -EINVAL;
}

static struct file_operations klife_device_fops = {
        .owner          = THIS_MODULE,
        .open           = klife_open,
        .release        = klife_release,
        .read           = klife_read,
        .write          = klife_write,
        .ioctl          = klife_ioctl,
};

static int __init klife_init(void) {
  int ret;

  ret = register_chrdev(KLIFE_MAJOR, "klife", &klife_device_fops);
  if (ret<0) {
    printk("Klife register_chrdev: %d\n", ret); 
  } else {
    printk("Loaded Klife\n");
  }

  return ret;
}

static void __exit klife_exit(void) {
  int ret;
  ret = unregister_chrdev(KLIFE_MAJOR, "klife");
  if (ret<0) {
    printk("Klife unregister_chrdev: %d\n", ret);
  } else {
    printk("Unloaded Klife\n");
  }
}

module_init(klife_init);
module_exit(klife_exit);

MODULE_AUTHOR("Elmar Pruesse");
MODULE_DESCRIPTION("Kernel Game of Life");
MODULE_LICENSE("GPL");

